"""
    DigiCloud DNS Domain, Record Service.
"""
from marshmallow.exceptions import ValidationError

from .base import Lister, ShowOne, Command
from ..error_handlers import CLIError
from .. import schemas


class ListDomain(Lister):
    """List dns domains"""
    schema = schemas.DNSDomainList(many=True)

    def get_data(self, parsed_args):
        domains = self.app.session.get('/dns/domains')
        return domains


class CreateDomain(ShowOne):
    """Create Domain"""
    schema = schemas.DNSDomainDetails()

    def get_parser(self, prog_name):
        parser = super().get_parser(prog_name)
        parser.add_argument(
            'name',
            metavar='<name>',
            help='Domain name'
        )
        return parser

    def get_data(self, parsed_args):
        payload = {
            'name': parsed_args.name,
        }
        domain = self.app.session.post('/dns/domains', payload)
        return domain


def t_or_f_or_n(arg):
    upper_arg = str(arg).upper()
    if 'TRUE'.startswith(upper_arg):
        return True
    elif 'FALSE'.startswith(upper_arg):
        return False
    else:
        return None


def get_record_schema(data):
    schema_map = {
        "A": schemas.ARecordDetailsSchema(),
        "TXT": schemas.TXTRecordDetailsSchema(),
    }
    return schema_map.get(data["type"], None)


class ListRecord(Lister):
    """List DNS records"""
    schema = schemas.RecordListSchema(many=True)

    def get_parser(self, prog_name):
        parser = super().get_parser(prog_name)
        parser.add_argument(
            '--domain',
            metavar='<domain>',
            help='DNS domain Name or ID',
        )
        return parser

    def get_data(self, parsed_args):
        domains = self.app.session.get('/dns/domains/{}/records'.format(parsed_args.domain))
        return domains


class ShowRecord(ShowOne):
    """Show domain details."""

    def get_parser(self, prog_name):
        parser = super().get_parser(prog_name)
        parser.add_argument(
            'record',
            metavar='<record>',
            help='DNS record ID',
        )
        parser.add_argument(
            '--domain',
            metavar='<domain>',
            help='DNS domain ID',
        )
        return parser

    def get_data(self, parsed_args):
        uri = '/dns/domains/{}/records/{}'.format(parsed_args.domain, parsed_args.record)
        record = self.app.session.get(uri)
        self.schema = get_record_schema(record)
        return record


class DeleteRecord(Command):
    """Delete record."""

    def get_parser(self, prog_name):
        parser = super().get_parser(prog_name)
        parser.add_argument(
            'record',
            metavar='<record>',
            help='DNS record ID',
        )
        parser.add_argument(
            '--domain',
            metavar='<domain>',
            help='DNS domain Name or ID',
        )
        return parser

    def take_action(self, parsed_args):
        uri = '/dns/domains/{}/records/{}'.format(parsed_args.domain, parsed_args.record)
        self.app.session.delete(uri)


class UpdateRecord(ShowOne):
    """Update record."""

    def get_parser(self, prog_name):
        parser = super().get_parser(prog_name)
        parser.add_argument(
            'record',
            metavar='<record>',
            help='DNS record ID',
        )
        parser.add_argument(
            '--domain',
            metavar='<domain>',
            help='DNS domain Name or ID.',
        )
        parser.add_argument(
            '--name',
            metavar='<name>',
            help='Record name.',
            required=False,
        )
        parser.add_argument(
            '--ttl',
            metavar='<ttl>',
            help='Time to live.',
            choices=("2m", "10m", "30m", "1h", "3h", "10h",),
            required=False,
        )
        parser.add_argument(
            '--ip-address',
            metavar='<ip-address>',
            help='IP Address.',
            required=False,
        )

        parser.add_argument(
            '--content',
            metavar='<content>',
            help='Content.',
            required=False,
        )
        parser.add_argument(
            '--port',
            metavar='<port>',
            help='Port.',
            required=False,
            type=int,
        )
        parser.add_argument(
            '--weight',
            metavar='<weight>',
            help='Weight.',
            required=False,
            type=int,
        )
        parser.add_argument(
            '--proto',
            metavar='<proto>',
            help='Proto.',
            required=False,
            choices=("_tcp", "_udp", "_tls"),
        )
        parser.add_argument(
            '--service',
            metavar='<service>',
            help='Service.',
            required=False,
        )
        parser.add_argument(
            '--target',
            metavar='<target>',
            help='Target.',
            required=False,
        )
        parser.add_argument(
            '--mail-server',
            metavar='<mail_server>',
            help='Mail server.',
            required=False,
        )
        parser.add_argument(
            '--priority',
            metavar='<priority>',
            help='Priority.',
            required=False,
            type=int,
        )
        return parser

    def get_data(self, parsed_args):
        uri = '/dns/domains/{}/records/{}'.format(parsed_args.domain, parsed_args.record)
        payload = {
            key: value for key, value in vars(parsed_args).items()
            if value is not None and key in schemas.RecordListSchema.available_keys
        }
        record = self.app.session.patch(uri, payload)
        self.schema = get_record_schema(record)
        return record


class CreateRecord(ShowOne):
    """Create Record"""

    def get_parser(self, prog_name):
        parser = super().get_parser(prog_name)
        parser.add_argument(
            '--domain',
            metavar='<domain>',
            help='DNS domain Name or ID.',
            required=True,
        )
        parser.add_argument(
            '--name',
            metavar='<name>',
            help='Record name.',
            required = True,
        )
        parser.add_argument(
            '--type',
            metavar='<type>',
            help='Record type.',
            choices=("A", "TXT", "CNAME", "MX", "SRV",),
            required=True,
        )
        parser.add_argument(
            '--ttl',
            metavar='<ttl>',
            help='Time to live.',
            choices=("2m", "10m", "30m", "1h", "3h", "10h",),
            required=True,
        )
        parser.add_argument(
            '--ip-address',
            metavar='<ip-address>',
            help='IP Address.',
            required=False,
        )

        parser.add_argument(
            '--content',
            metavar='<content>',
            help='Content.',
            required=False,
        )
        parser.add_argument(
            '--port',
            metavar='<port>',
            help='Port.',
            required=False,
            type=int,
        )
        parser.add_argument(
            '--weight',
            metavar='<weight>',
            help='Weight.',
            required=False,
            type=int,
        )
        parser.add_argument(
            '--proto',
            metavar='<proto>',
            help='Proto.',
            required=False,
            choices=("_tcp", "_udp", "_tls"),
        )
        parser.add_argument(
            '--service',
            metavar='<service>',
            help='Service.',
            required=False,
        )
        parser.add_argument(
            '--target',
            metavar='<target>',
            help='Target.',
            required=False,
        )
        parser.add_argument(
            '--mail-server',
            metavar='<mail_server>',
            help='Mail server.',
            required=False,
        )
        parser.add_argument(
            '--priority',
            metavar='<priority>',
            help='Priority.',
            required=False,
            type=int,
        )
        return parser

    def get_data(self, parsed_args):
        payload = {
            key: value for key, value in vars(parsed_args).items()
            if value is not None and key in schemas.RecordListSchema.available_keys
        }
        try:
            payload = self._get_record_type_schema(payload["type"])().load(payload)
        except ValidationError as e:
            raise CLIError(self._handle_validation_error(payload["type"], e))
        record = self.app.session.post('/dns/domains/{}/records'.format(parsed_args.domain), payload)
        self.schema = get_record_schema(record)
        return record

    @staticmethod
    def _handle_validation_error(record_type: str, e: ValidationError):
        errors = []
        for key, value in e.messages.items():
            error = "".join(value)
            if error == "Unknown field.":
                msg = "can not use --{} with {} record type.".format(key, record_type)
            elif error == "Missing data for required field.":
                msg = "--{} is required by {} record type.".format(
                    key.replace("_", "-"),
                    record_type
                )
            else:
                raise NotImplementedError
            errors.append(dict(
                msg=msg
            ))
        return errors

    @staticmethod
    def _get_record_type_schema(record_type: str):
        record_schema_map = {
            "A": schemas.ARecordDetailsSchema,
            "TXT": schemas.TXTRecordDetailsSchema,
            "MX": schemas.MXRecordDetailsSchema,
            "CNAME": schemas.CNAMERecordDetailsSchema,
            "SRV": schemas.SRVRecordDetailsSchema,
        }
        return record_schema_map[record_type]



"""
    DigiCloud load balancer Service.
"""
import ipaddress

from .base import Lister, ShowOne, Command
from .. import schemas
from ..error_handlers import CLIError


def get_load_balancer_details(session, load_balancer):
    uri = '/load-balancers/%s' % load_balancer
    load_balancer = session.get(uri)
    return load_balancer


################################################
# Load Balancer                                #
################################################
class ListLoadBalancer(Lister):
    """List load balancers"""
    schema = schemas.LoadBalancerList(many=True)

    def get_data(self, parsed_args):
        load_balancer_list = self.app.session.get('/load-balancers')
        return load_balancer_list


class ShowLoadBalancer(ShowOne):
    """Show load balancer details."""
    schema = schemas.LoadBalancerDetail()

    def get_parser(self, prog_name):
        parser = super().get_parser(prog_name)
        parser.add_argument(
            'load_balancer',
            metavar='<load-balancer>',
            help='Load Balancer name or ID',
        )
        return parser

    def get_data(self, parsed_args):
        uri = '/load-balancers/%s' % parsed_args.load_balancer
        load_balancer = self.app.session.get(uri)
        return load_balancer


class DeleteLoadBalancer(Command):
    """Delete load balancer."""

    def get_parser(self, prog_name):
        parser = super().get_parser(prog_name)
        parser.add_argument(
            'load_balancer',
            metavar='<load-balancer>',
            help='Load Balancer name or ID',
        )
        return parser

    def take_action(self, parsed_args):
        uri = '/load-balancers/%s' % parsed_args.load_balancer
        self.app.session.delete(uri)


class UpdateLoadBalancer(ShowOne):
    """Update load balancer."""
    schema = schemas.LoadBalancerDetail()

    def get_parser(self, prog_name):
        parser = super().get_parser(prog_name)
        parser.add_argument(
            'load_balancer',
            metavar='<load-balancer>',
            help='Load Balancer name or ID',
        )
        parser.add_argument(
            '--name',
            metavar='<Name>',
            help='New name for load balancer.'
        )
        parser.add_argument(
            '--description',
            metavar='<description>',
            help='Load balancer description.',
        )
        parser.add_argument(
            '--delay',
            metavar='<delay>',
            help='Health Check delay.',
            type=int,
        )
        parser.add_argument(
            '--timeout',
            metavar='<timeout>',
            help='Health Check timeout.',
            type=int,
        )
        parser.add_argument(
            '--max-retries',
            metavar='<max-retries>',
            help='Health Check max retries.',
            type=int,
        )
        parser.add_argument(
            '--path',
            metavar='<path>',
            help='Health Check path',
        )
        parser.add_argument(
            '--http-version',
            metavar='<http-version>',
            help='Health Check http version',
            type=float,
            choices=[1.0, 1.1],
        )
        parser.add_argument(
            '--http-method',
            metavar='<http-method>',
            help='Health Check http method',
            choices=['GET', 'POST'],
        )
        parser.add_argument(
            '--algorithm',
            metavar='<algorithm>',
            help='Backend algorithm',
            choices=['round_robin', 'source_ip'],
        )
        return parser

    def get_data(self, parsed_args):
        uri = '/load-balancers/%s' % parsed_args.load_balancer
        payload = {}

        health_check_payload = {}
        if parsed_args.delay:
            health_check_payload['delay'] = parsed_args.delay
        if parsed_args.timeout:
            health_check_payload['timeout'] = parsed_args.timeout
        if parsed_args.max_retries:
            health_check_payload['max_retries'] = parsed_args.max_retries
        if parsed_args.path:
            health_check_payload['path'] = parsed_args.path
        if parsed_args.http_version:
            health_check_payload['http_version'] = parsed_args.http_version
        if parsed_args.http_method:
            health_check_payload['http_method'] = parsed_args.http_method

        if parsed_args.name:
            payload['name'] = parsed_args.name
        if parsed_args.description:
            payload['description'] = parsed_args.description

        if health_check_payload:
            payload["application"] = {
                "backends": [{
                    "health_check": health_check_payload
                }]
            }

        if parsed_args.algorithm:
            if payload.get("application"):
                payload["application"]["backends"][0]["algorithm"] = parsed_args.algorithm
            else:
                payload["application"] = {
                    "backends": [{
                        "algorithm": parsed_args.algorithm
                    }]
                }

        if not payload:
            raise CLIError([dict(
                msg="At least one of --delay or"
                    " --timeout or"
                    " --max-retries or"
                    " --path or"
                    " --http-version or"
                    " --http-method or"
                    " --algorithm or"
                    " --name or"
                    " --description is necessary"
            )])
        load_balancer = self.app.session.patch(uri, payload)
        return load_balancer


class CreateLoadBalancer(ShowOne):
    """Create Load Balancer"""
    schema = schemas.LoadBalancerDetail()

    def get_parser(self, prog_name):
        parser = super().get_parser(prog_name)
        parser.add_argument(
            '--name',
            metavar='<Name>',
            help='New name for load balancer.',
            required=True,
        )
        parser.add_argument(
            '--description',
            metavar='<description>',
            help='Load balancer description.',
        )
        return parser

    def get_data(self, parsed_args):
        payload = {
            'name': parsed_args.name,
        }
        if parsed_args.description:
            payload['description'] = parsed_args.description
        load_balancer = self.app.session.post('/load-balancers', payload)
        return load_balancer


################################################
# Application                                  #
################################################
class CreateApplication(ShowOne):
    """Create Application"""
    schema = schemas.ApplicationDetail()

    def get_parser(self, prog_name):
        parser = super().get_parser(prog_name)
        parser.add_argument(
            'load_balancer',
            metavar='<load-balancer>',
            help='Load Balancer name or ID',
        )
        parser.add_argument(
            '--port',
            metavar='<port>',
            help='Frontend port.',
            type=int,
            required=True,
        )
        parser.add_argument(
            '--algorithm',
            metavar='<algorithm>',
            help='Backend algorithm',
            choices=['round_robin', 'source_ip'],
            required=True,
        )
        return parser

    def _check_arg_validity(self, parsed_args):
        rules = [
            (
                parsed_args.port < 1 or parsed_args.port > 65535,
                "port should be in range 1 - 65535",
            ),
        ]
        errors = []
        for is_invalid, err_msg in rules:
            if is_invalid:
                errors.append(dict(msg=err_msg))
        if errors:
            raise CLIError(errors)

    def get_data(self, parsed_args):
        self._check_arg_validity(parsed_args)
        payload = {
            "backends": [
                {
                    'algorithm': parsed_args.algorithm,
                },
            ],
            "frontends": [
                {
                    'port': parsed_args.port,
                },
            ],
        }
        uri = '/load-balancers/%s/applications' % parsed_args.load_balancer
        application = self.app.session.post(uri, payload)
        return application


################################################
# Health Check                                 #
################################################
class DeleteHealthCheck(Command):
    """Delete Health Check."""

    def get_parser(self, prog_name):
        parser = super().get_parser(prog_name)
        parser.add_argument(
            'load_balancer',
            metavar='<load-balancer>',
            help='Load Balancer name or ID',
        )
        return parser

    def take_action(self, parsed_args):
        load_balancer = get_load_balancer_details(self.app.session, parsed_args.load_balancer)
        uri = '/load-balancers/%s/applications/%s/backends/%s/health-checks/%s' % (
            parsed_args.load_balancer,
            load_balancer['application']['id'],
            load_balancer['application']['backends'][0]["id"],
            load_balancer['application']['backends'][0]['health_check']['id'],
        )
        self.app.session.delete(uri)


class CreateHealthCheck(ShowOne):
    """Create Health Check"""
    schema = schemas.HealthCheckDetail()

    def get_parser(self, prog_name):
        parser = super().get_parser(prog_name)
        parser.add_argument(
            'load_balancer',
            metavar='<load-balancer>',
            help='Load Balancer name or ID',
        )
        parser.add_argument(
            '--delay',
            metavar='<delay>',
            help='Health Check delay.',
            type=int,
            required=True,
        )
        parser.add_argument(
            '--timeout',
            metavar='<timeout>',
            help='Health Check timeout.',
            type=int,
            required=True
        )
        parser.add_argument(
            '--max-retries',
            metavar='<max-retries>',
            help='Health Check max retries.',
            type=int,
            required=True,
        )
        parser.add_argument(
            '--path',
            metavar='<path>',
            help='Health Check path',
        )
        parser.add_argument(
            '--http-version',
            metavar='<http-version>',
            help='Health Check http version',
            type=float,
            choices=[1.0, 1.1],
        )
        parser.add_argument(
            '--http-method',
            metavar='<http-method>',
            help='Health Check http method',
            choices=['GET', 'POST'],
        )
        return parser

    def _check_arg_validity(self, parsed_args):
        rules = [
            (
                parsed_args.delay < parsed_args.timeout,
                "timeout must be lower than and equal to delay",
            ),
        ]
        errors = []
        for is_invalid, err_msg in rules:
            if is_invalid:
                errors.append(dict(msg=err_msg))
        if errors:
            raise CLIError(errors)

    def get_data(self, parsed_args):
        self._check_arg_validity(parsed_args)
        load_balancer = get_load_balancer_details(self.app.session, parsed_args.load_balancer)
        uri = '/load-balancers/%s/applications/%s/backends/%s/health-checks' % (
            parsed_args.load_balancer,
            load_balancer['application']['id'],
            load_balancer['application']['backends'][0]["id"],
        )
        payload = {
            'delay': parsed_args.delay,
            'timeout': parsed_args.timeout,
            'max_retries': parsed_args.max_retries,
            'path': parsed_args.path,
            'http_version': parsed_args.http_version,
            'http_method': parsed_args.http_method,
        }
        health_check = self.app.session.post(uri, payload)
        return health_check


################################################
# Backend Member                               #
################################################
class DeleteBackendMember(Command):
    """Delete backend member."""

    def get_parser(self, prog_name):
        parser = super().get_parser(prog_name)
        parser.add_argument(
            'load_balancer',
            metavar='<load-balancer>',
            help='Load Balancer name or ID',
        )
        parser.add_argument(
            'member',
            metavar='<member>',
            help='Member ID',
        )
        return parser

    def take_action(self, parsed_args):
        load_balancer = get_load_balancer_details(self.app.session, parsed_args.load_balancer)
        uri = '/load-balancers/%s/applications/%s/backends/%s/members/%s' % (
            parsed_args.load_balancer,
            load_balancer['application']['id'],
            load_balancer['application']['backends'][0]["id"],
            parsed_args.member,
        )
        self.app.session.delete(uri)


class AddBackendMember(Lister):
    """Add Backend member"""
    schema = schemas.BackendMemberDetail(many=True)

    def get_parser(self, prog_name):
        parser = super().get_parser(prog_name)
        parser.add_argument(
            'load_balancer',
            metavar='<load-balancer>',
            help='Load Balancer name or ID',
        )
        parser.add_argument(
            '--resource-id',
            metavar='<resource-id>',
            help='Backend resource id.',
            required=True,
        )
        parser.add_argument(
            '--ip-address',
            metavar='<ip-address>',
            help='Backend member ip address.',
            required=True,
        )
        parser.add_argument(
            '--port',
            metavar='<port>',
            help='Backend member port.',
            type=int,
            required=True,
        )
        return parser

    def _check_ip_address(self, value):
        try:
            ipaddress.IPv4Address(value)
            return False
        except (ipaddress.AddressValueError, ipaddress.NetmaskValueError):
            return True

    def _check_arg_validity(self, parsed_args):
        rules = [
            (
                parsed_args.port and (parsed_args.port < 1 or parsed_args.port > 65535),
                "port should be in range 1 - 65535",
            ),
            (
                self._check_ip_address(parsed_args.ip_address),
                "Not a valid ipv4 address"
            )
        ]
        errors = []
        for is_invalid, err_msg in rules:
            if is_invalid:
                errors.append(dict(msg=err_msg))
        if errors:
            raise CLIError(errors)

    def get_data(self, parsed_args):
        self._check_arg_validity(parsed_args)
        load_balancer = get_load_balancer_details(self.app.session, parsed_args.load_balancer)
        uri = '/load-balancers/%s/applications/%s/backends/%s/members' % (
            parsed_args.load_balancer,
            load_balancer['application']['id'],
            load_balancer['application']['backends'][0]["id"],
        )
        payload = [{
            'resource_id': parsed_args.resource_id,
            'ip_address': parsed_args.ip_address,
            'port': parsed_args.port,
        }]
        backend_member = self.app.session.post(uri, payload)
        return backend_member



from .base import Lister, ShowOne
from .. import schemas


class QuotaRequestList(Lister):
    """List your namespace quota"""
    schema = schemas.QuotaRequest(many=True)

    def get_data(self, parsed_args):
        return self.app.session.get('/quota-requests')[:15]


class RequestMoreQuota(ShowOne):
    """Submit a request for more quota"""
    schema = schemas.QuotaRequest()

    def get_parser(self, prog_name):
        parser = super(RequestMoreQuota, self).get_parser(prog_name)
        parser.add_argument(
            '--quota-id',
            required=True,
            metavar='<quota_id>',
            help='Quota ID for the request',
        )

        parser.add_argument(
            '--value',
            required=True,
            metavar='<value>',
            help='Your desire value'
        )

        return parser

    def get_data(self, parsed_args):
        return self.app.session.post(
            '/quota-requests',
            {'quota_id': parsed_args.quota_id,
             'required_quota': parsed_args.value,
             'note': "Requested by CLI"}
        )
